% same as hierarchy.erl, but balance the load on different slaves 

% assumes the existence of nodes slave1, slave2, accessible from
% current node (master)
% ex. 
%     erl -sname master -setcookie "a"
%     erl -sname slave1 -setcookie "a"
%     erl -sname slave2 -setcookie "a"
% each having the full code available (could be better split in files
% ... exercise)

-module(distributed).
-export([start/0, servant/3, server/3,client/1]).

servant(F, Args, Server) ->
    % randomly fails
    case rand:uniform(4) of
	1 ->     
	    Server ! {answ, {self(), F(Args)}},
	    erlang:display("servant worked");
	_ ->
	    exit(went_wrong)
    end.



% Creator     = identita' del creatore, solo lui puo' chiudere  
% WaitingList = Lista di richieste lanciate e non ancora ricevute 
%               del tipo {Servant,Client,F,Args}
% --> CHANGED <--
% Slaves =      list of slaves, first in the list is next to use, and then
%               reinserted as last

server(Creator, WaitingList, Slaves) ->
    % timer:sleep(1000),

    % Server traps the exit signals
    process_flag(trap_exit, true),
    
    receive
        % execution requests (Pid of the client, F function to be
        % executed on Args), establishes a (trapped) link!
        {req, {Client, F, Args}} ->            
	    % get the next slave
	    [Slave|SlavesRest] = Slaves,
	    % spawn and link at the same time (othewise spawn followed
	    % by link could be non-atomic), on the Slave machine, 
	    % that then is put at the end of the queue
	    io:format("Server spawning on slave ~w~n", [Slave]),
            Servant = spawn_link(Slave, distributed, servant, [F, Args, self()]),	    
	    server(Creator, 
		   [{Servant,Client,F,Args} | WaitingList], 
		   SlavesRest++[Slave]);

	% normal termination message from servant
	{answ, {Servant, Result}} ->
	    {_,Client,F,Args} = lists:keyfind(Servant,1,WaitingList),
	    Client ! {answ, {Client, F,Args}, Result},
	    server(Creator, WaitingList--[{Servant,Client,F,Args}], Slaves);
	

	% error message from callback
	{'EXIT', Servant, went_wrong } ->
	    % get the next slave
	    [Slave|SlavesRest] = Slaves,

	    {_,Client,F,Args} = lists:keyfind(Servant,1,WaitingList),
	    io:format("Servant ~w went wrong, retrying on slave ~w ...~n", [Servant,Slave]),
	    NewServant = spawn_link(Slave,distributed, servant, [F, Args, self()]),
	    server(Creator, (WaitingList--[{Servant,Client,F,Args}])++[{NewServant,Client,F,Args}],SlavesRest++[Slave]);
	
	% only the creator can stop the server
        {Creator, stop} ->
            io:format("Server stopping ...~n", []),
	    exit(normal) ;
	
        {Pid, stop} ->
            io:format("Process \"~w\" not allowed to stop the server ...~n", [Pid]),
            server(Creator,WaitingList,Slaves)
		
    end.

% facilities for outputting the result
out_result(Name, fail) ->
    io:format("Timed out waiting for ~w~n", [Name]);
out_result(Name, Result) ->
    io:format("result of ~w is ~B~n", [Name, Result]).
    

client(Server) ->
    % first message
    Msg1 = {self(), fun([X,Y]) -> X*Y end, [1,2]},
    Server ! {req, Msg1},

    receive 
         {answ, Msg1, R1} -> 
    	    out_result(product,R1)
    after 
	10 ->
	    out_result(product,fail)
    end,
    
    Server ! {self(), stop},
    

    % second message (receive order could be swapped)
    % first message
    Msg2 = {self(), fun([X,Y]) -> X+Y end, [1,2]},
    Server ! {req, Msg2},

    receive 
        {answ, Msg2, R2} -> 
	    out_result(sum,R2)
     after 
      	10 ->
      	    out_result(sum,fail)
    end.
    
% build full node name given the short name and the host
build_node_name (Sname, Host) ->
    Sname ++ "@" ++ Host.


start() ->
    % list of slaves
    SlavesSnames = ["slave1","slave2"],
    % could get get existing Erlang nodes via
    % Nodes = net_adm:names().

    % localhost (trick for removing a trailing ".local" from host name)
    Host = re:replace(net_adm:localhost(),"\\..*","",[{return,list}]),
    
    
    Slaves = lists:map(fun(Sname) -> list_to_atom(build_node_name(Sname,Host)) end, SlavesSnames),


    % connect to slaves
    lists:map(fun(Node) -> net_kernel:connect_node(Node) end, Slaves),


    % start the server
    Server = spawn(distributed, server, [self(),[],Slaves]),

    % to register and use a public name one would need to refer also
    % to the node hosting the name (simplified using pids)
    % register(pserver, Server),

    % accessible to some client, without getting the pid, but the node
    % name must be used
    spawn(distributed, client, [Server]),

    % wait a bit and stops the server
    timer:sleep(1000),
    Server ! {self(),stop}.
