% supervisor receiving requests that are delegated to concurrent servants
% servants can fail, if so they are restarted
-module(hierarchy).
-export([start/0, servant/3, server/2, client/0]).


% Given a function and some arguments, executes the functions on the arguments
% randomly fails (75% of the times)
servant(F, Args, Server) ->    
    case rand:uniform(4) of
                    
        % regular execution, notify the server providing the result
        1 ->
            Server ! {answ, {self(), F(Args)}};

        % failure
        _ ->
            exit(went_wrong)
    end.


% The server keeps in its state
%
% - Creator: the pid of the creator
%
% - WaitingList = list of requests being processed of the kind 
%   {Servant,Client,F,Args} including Servant's pid, client's pid, request data

server(Creator, WaitingList) ->
    % timer:sleep(1000),

    % Supervisor process: traps the exit signals
    process_flag(trap_exit, true),
    
    receive
        % execution requests (Pid of the client, F function to be
        % executed on Args), establishes a (trapped) link!
        {req, {Client, F, Args}} ->
            % spawn and link at the same time (othewise spawn followed
            % by link could be non-atomic)
            Servant = spawn_link(hierarchy, servant, [F, Args, self()]),            
            server(Creator, [{Servant,Client,F,Args} | WaitingList]);

        % normal termination message from servant
        {answ, {Servant, Result}} ->
            {_,Client,F,Args} = lists:keyfind(Servant,1,WaitingList),
            Client ! {answ, {Client, F,Args}, Result},
            server(Creator, WaitingList--[{Servant,Client,F,Args}]);
        

        % error message from servant
        {'EXIT', Servant, went_wrong } ->
            {_,Client,F,Args} = lists:keyfind(Servant,1,WaitingList),
            io:format("Servant ~w went wrong, retrying ...~n", [Servant]),
            NewServant = spawn_link(hierarchy, servant, [F, Args, self()]),
            server(Creator, (WaitingList--[{Servant,Client,F,Args}])++[{NewServant,Client,F,Args}] );
        
        % only the creator can stop the server
        {Creator, stop} ->
            io:format("Server stopping ...~n", []),
            exit(normal) ;
        
        {Pid, stop} ->
            io:format("Process \"~w\" not allowed to stop the server ...~n", [Pid]),
            server(Creator,WaitingList)
                
    end.

% facilities for outputting the result
out_result(Name, fail) ->
    io:format("Timed out waiting for ~w~n", [Name]);
out_result(Name, Result) ->
    io:format("result of ~w is ~B~n", [Name, Result]).
    

client() ->
    % first message
    Msg1 = {self(), fun([X,Y]) -> X*Y end, [1,2]},
    pserver ! {req, Msg1},

    receive 
         {answ, Msg1, R1} -> 
            out_result(product,R1)
    after 
        10 ->
            out_result(product,fail)
    end,
    
    pserver ! {self(), stop},
    

    % second message (receive order could be swapped)
    % first message
    Msg2 = {self(), fun([X,Y]) -> X+Y end, [1,2]},
    pserver ! {req, Msg2},

    receive 
        {answ, Msg2, R2} -> 
            out_result(sum,R2)

      after 
        10 ->
            out_result(sum,fail)
    end.
    
    

start() ->

    % the server is registered
    Server = spawn(hierarchy, server, [self(),[]]),

    % public name
    register(pserver, Server),

    % accessible to some client, without getting the pid
    spawn(hierarchy, client, []),

    % wait a bit and stops the server
    timer:sleep(1000),
    pserver ! {self(),stop}.
