-module(mulGenServer).

% declare that the gen_server behaviour is implemented
-behaviour(gen_server).

-export([go/0,client/0]).
-export([start/0, mul/2, stop/0]).
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2]).


%%% INTERFACE

% Create the server, registered locally as multiplier, calling init
% with parameter self() (the pid of the creator)
start() ->
    gen_server:start({local, multiplier}, ?MODULE, [self()], []).

% multiplication: synchronous call
mul(X,Y) ->
    gen_server:call(multiplier, {mul, X, Y}).

% stop request, asynchronous call passing the pid of the caller
% (better implemented as terminate message, just to have an example of
% cast)
stop() ->
    gen_server:cast(multiplier, {stop, self()}).



%%% CALLBACKS

% initialization: establish the initial state
init([Creator]) ->
    {ok, [Creator]}.

% multiplication handle
% IN:   message, sender, server state
% OUT:  reply atom, reply content, new state
handle_call({mul,X,Y}, _From, [Creator]) ->
    {reply, X*Y, [Creator]}.

% stop handle
handle_cast({stop, From}, [Creator]) ->
    if From =:= Creator ->
	    {stop, normal, [Creator]};
        From =/= Creator ->
	    io:format("Invalid shutdown req (pid ~w)~n",[From]),    
	    {noreply, [Creator]}
    end.

% handling termination
terminate(normal, [Creator]) ->
    io:format("Server created by: ~w properly terminated~n",[Creator]).  

% other messages
handle_info(Msg, [Creator]) ->
    io:format("Unexpected message: ~p~n",[Msg]),
    {noreply, [Creator]}.

    

%%% CLIENT

client() ->
    % first 
    A1=2,
    A2=2,
    Result = mulGenServer:mul(A1,A2),
    io:format("Product of ~B * ~B is ~B~n", [A1, A2, Result]),


    mulGenServer:stop().
    
    

go() ->
    % the server is started
    mulGenServer:start(),

    % and then the client
    spawn(?MODULE, client, []),

    % wait a bit and stops the server
    timer:sleep(1000),
    mulGenServer:stop().

